#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Author: Luis Bonah
# Description : CLI tool for showing parameter uncertainties

import os
import argparse
import pyckett

def report():
    parser = argparse.ArgumentParser(prog="Summarize an analysis")

    parser.add_argument("linfile", type=str, help="Filename of the .lin file")
    parser.add_argument(
        "catfile", type=str, nargs="?", help="Filename of the .cat file"
    )

    parser.add_argument(
        "--noblends",
        action="store_false",
        default=None,
        dest='blends',
        help="Skip blend treatment",
    )

    parser.add_argument(
        "--noq",
        type=int,
        nargs=1,
        help="Specify number of quantum numbers",
    )

    args = parser.parse_args()


    linfname = args.linfile
    base, ext = os.path.splitext(linfname)
    if not ext:
        linfname = linfname + ".lin"

    catfname = args.catfile if args.catfile else linfname.replace(".lin", ".cat")

    lin = pyckett.lin_to_df(linfname, sort=False)
    try:
        cat = pyckett.cat_to_df(catfname, sort=False)
    except FileNotFoundError:
        cat = None
        pass

    report, _ = pyckett.create_report(lin, cat, blends=args.blends, noq=args.noq)
    print(report)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Author: Luis Bonah
# Description : CLI tool for creating separate fits from global fit

import os
import argparse
import pyckett
import copy
from concurrent.futures import ThreadPoolExecutor


def separatefits():
    parser = argparse.ArgumentParser(prog="Separate Fits")

    parser.add_argument("linfile", type=str, help="Filename of the .lin file")
    parser.add_argument(
        "parfile", type=str, nargs="?", help="Filename of the .par file"
    )

    parser.add_argument(
        "--parupdate",
        action="store_true",
        help="Reset .par parametes NPAR, NLINE, THRESH",
    )
    parser.add_argument(
        "--stateqn", default=4, type=int, help="Quantum number index for state"
    )
    parser.add_argument("--forcesingles", action="store_true", help="Force single fits")
    parser.add_argument(
        "--keepqns", action="store_true", help="Keep the original state qns"
    )

    args = parser.parse_args()

    linfname = args.linfile
    base, ext = os.path.splitext(linfname)
    if not ext:
        linfname = linfname + ".lin"

    lin = pyckett.lin_to_df(linfname, sort=False)
    par = pyckett.parvar_to_dict(
        args.parfile if args.parfile else linfname.replace(".lin", ".par")
    )

    VIB_DIGITS = pyckett.get_vib_digits(par)
    ALL_STATES = pyckett.get_all_states(VIB_DIGITS)

    if args.parupdate:
        par.update(pyckett.PARUPDATE)

    if not os.path.isdir("separatefits"):
        os.mkdir("separatefits")

    qnu, qnl = f"qnu{args.stateqn:1.0f}", f"qnl{args.stateqn:1.0f}"

    params = []
    states_sets = []
    for param in par["PARAMS"]:
        parsed_id = pyckett.parse_param_id(param[0], VIB_DIGITS)
        v1, v2 = parsed_id["v1"], parsed_id["v2"]

        params.append([v1, v2, param])

        if v1 == v2:
            contains_v = [v1 in states for states in states_sets]
            if not any(contains_v):
                states_sets.append(set((v1,)))
        elif not args.forcesingles:
            popped_sets = []

            for i, states in enumerate(states_sets):
                if v1 in states or v2 in states:
                    popped_sets.append(i)

            popped_sets = [states_sets.pop(i) for i in popped_sets[::-1]]
            new_set = set((v1, v2)).union(*popped_sets)

            states_sets.append(new_set)

    states_without_parameters = set(lin[qnu].unique()) & set(lin[qnl].unique())
    for v in states_without_parameters:
        if all([v not in states_set for states_set in states_sets]):
            states_sets.append(set((v,)))

    # @Luis: Think about states that are only in lin file

    def worker(i):
        states = sorted(states_sets[i])
        states_identifier = "_".join([f"{state:03.0f}" for state in states])
        filename = f"State_{states_identifier}"

        # @Luis: This still needs some consideration in the future
        ##########################################################################################
        # tmp_lin = lin.query(f"{qnu} in @states and {qnl} in @states").copy()
        ##########################################################################################
        # tmp_lin = lin.query(f"({qnu} in @states and {qnl} in @states) or (error < 0 and ({qnu} in @states or {qnl} in @states))").copy()
        tmp_lin = lin.query(
            f"({qnu} in @states and {qnl} in @states) or (error < 0 and {qnu} in @states)"
        ).copy()
        for qn in (qnu, qnl):
            states.extend(tmp_lin[qn].unique())
        states = sorted(set(states))
        ##########################################################################################

        filter_states = lambda x: (x[0] in states and x[1] in states) or (
            x[0] == ALL_STATES and x[1] == ALL_STATES
        )
        tmp_params = [x[-1].copy() for x in params if filter_states(x)]
        tmp_par = copy.deepcopy(par)

        tmp_par["STATES"] = [
            state for state in tmp_par["STATES"] if state.get("NVIB") in states
        ]
        if len(tmp_par["STATES"]):
            tmp_par["STATES"][-1]["VSYM"] = 1
        elif tmp_par["VSYM"] < 0:
            tmp_par["VSYM"] = 0

        if not args.keepqns:
            # Update vsym to match the new quantum numbers
            vsym = tmp_par["VSYM"]
            if vsym > 0:
                vsym_per_state = [int(x) for x in str(vsym)][::-1]
                vsym_per_state = [
                    str(vsym_per_state[x]) for x in states if x < len(vsym_per_state)
                ][::-1]

                new_vsym = "".join(vsym_per_state) if vsym_per_state else "0"
                tmp_par["VSYM"] = int(new_vsym)

            sign_nvib = -1 if tmp_par["NVIB"] < 0 else 1
            tmp_par["NVIB"] = max(2, len(states)) * sign_nvib
            new_vib_digits = pyckett.get_vib_digits(tmp_par)
            new_all_states = pyckett.get_all_states(new_vib_digits)

            translation_dict = {v: i for i, v in enumerate(states)}
            translation_dict[ALL_STATES] = new_all_states

            tmp_lin[qnu] = tmp_lin[qnu].replace(translation_dict)
            tmp_lin[qnl] = tmp_lin[qnl].replace(translation_dict)

            for param in tmp_params:
                parsed_id = pyckett.parse_param_id(param[0], VIB_DIGITS)
                parsed_id["v1"] = translation_dict[parsed_id["v1"]]
                parsed_id["v2"] = translation_dict[parsed_id["v2"]]
                param[0] = pyckett.format_param_id(parsed_id, new_vib_digits)

            for state in tmp_par["STATES"]:
                state["NVIB"] = translation_dict[state["NVIB"]]

        tmp_par["PARAMS"] = tmp_params
		# @Luis: Check here if we can summarize parameters
		# - for each unique parameter check if there are
			# - same parameter again
			# - corresponding global parameter

        params_are_global = []
        for param in tmp_par["PARAMS"]:
            parsed_id = pyckett.parse_param_id(param[0], new_vib_digits)
            v1, v2 = parsed_id["v1"], parsed_id["v2"]
            param_is_global = v1 == v2 == new_all_states
            params_are_global.append(param_is_global)

        if not all(params_are_global):
            for param, is_global in zip(tmp_par["PARAMS"], params_are_global):
                if is_global:
                    param[2] = pyckett.ZERO

        with open(os.path.join("separatefits", filename + ".lin"), "w+") as file:
            file.write(pyckett.df_to_lin(tmp_lin))

        with open(os.path.join("separatefits", filename + ".par"), "w+") as file:
            file.write(pyckett.dict_to_parvar(tmp_par))

        if len(tmp_lin):
            message = pyckett.run_spfit(filename, wd="separatefits")
            stats = pyckett.parse_fit_result(
                message,
                pyckett.parvar_to_dict(os.path.join("separatefits", filename + ".var")),
            )
        else:
            stats = {}
        stats["states"] = states
        stats["total_transitions"] = len(tmp_lin)
        stats["total_lines"] = len(tmp_lin["x"].unique())

        return stats

    with ThreadPoolExecutor() as executor:
        futures = {i: executor.submit(worker, i) for i in range(len(states_sets))}
        runs = [f.result() for f in futures.values()]

    header = (
        "States         |  RMS / kHz  |    WRMS    |  Rejected  |  Lines  |  Trans  "
    )
    print(header)
    print("-" * len(header))
    for results in sorted(runs, key=lambda x: x["states"][0]):
        states_identifier = "_".join([f"{state:03.0f}" for state in results["states"]])
        if "rms" in results:
            # print(f"States {states_identifier:15};   RMS {results['rms']*1000 :12.4f} kHz; Rejected lines {results['rejected_lines'] :7.0f} /{results['total_lines'] :7.0f}")
            wrms = float(results["wrms"])
            print(
                f"{states_identifier:15}|{results['rms']*1000 :12.4f} |{wrms:11.4f} |{results['rejected_lines'] :11.0f} |{results['total_lines'] :8.0f} |{results['total_transitions'] :8.0f} "
            )



# -*- coding: utf-8 -*-
#!/usr/bin/env python3

# Author: Luis Bonah
# Description : Tests for Pyckett library

import unittest
import pyckett
import numpy as np


class TestFormatting(unittest.TestCase):
    self.assertEqual(pyckett.format_(359, '2p'), 'Z9')
    self.assertEqual(pyckett.format_(360, '2p'), '**')
    self.assertEqual(pyckett.format_(12, '2p'), '12')
    self.assertEqual(pyckett.format_(-10, '2p'), 'a0')
    self.assertEqual(pyckett.format_(-270, '2p'), '**')
    self.assertEqual(pyckett.format_(9, '3p'), '  9')
    self.assertEqual(pyckett.pickett_int('Z9'), 359)
    self.assertEqual(pyckett.pickett_int('**'), 99)
    self.assertEqual(pyckett.pickett_int('12'), 12)
    self.assertEqual(pyckett.pickett_int('a0'), -10)
    self.assertEqual(pyckett.pickett_int('**'), 99)

class TestLabels(unittest.TestCase):
    def test_QNLABELS(self):
        self.assertEqual(pyckett.qnlabels_from_quanta(), pyckett.QNLABELS)

    def test_QNLABELS_ENERGY(self):
        self.assertEqual(pyckett.qnlabels_energy_from_quanta(), pyckett.QNLABELS_ENERGY)


class TestCatFormat(unittest.TestCase):
    def test_cat_dtypes(self):
        self.assertEqual(pyckett.cat_dtypes_from_quanta(), pyckett.cat_dtypes)
        self.assertEqual(
            pyckett.cat_dtypes_from_quanta(), pyckett.cat_dtypes_from_quanta(6)
        )

        self.assertEqual(
            pyckett.cat_dtypes_from_quanta(3),
            {
                "x": np.float64,
                "error": np.float64,
                "y": np.float64,
                "degfreed": pyckett.pickett_int,
                "elower": np.float64,
                "usd": pyckett.pickett_int,
                "tag": pyckett.pickett_int,
                "qnfmt": pyckett.pickett_int,
                **{
                    f"qn{ul}{i+1}": pyckett.pickett_int
                    for ul in ("u", "l")
                    for i in range(6)
                },
            },
        )

        self.assertEqual(
            pyckett.cat_dtypes_from_quanta(6),
            {
                "x": np.float64,
                "error": np.float64,
                "y": np.float64,
                "degfreed": pyckett.pickett_int,
                "elower": np.float64,
                "usd": pyckett.pickett_int,
                "tag": pyckett.pickett_int,
                "qnfmt": pyckett.pickett_int,
                **{
                    f"qn{ul}{i+1}": pyckett.pickett_int
                    for ul in ("u", "l")
                    for i in range(6)
                },
            },
        )

        self.assertEqual(
            pyckett.cat_dtypes_from_quanta(8),
            {
                "x": np.float64,
                "error": np.float64,
                "y": np.float64,
                "degfreed": pyckett.pickett_int,
                "elower": np.float64,
                "usd": pyckett.pickett_int,
                "tag": pyckett.pickett_int,
                "qnfmt": pyckett.pickett_int,
                **{
                    f"qn{ul}{i+1}": pyckett.pickett_int
                    for ul in ("u", "l")
                    for i in range(8)
                },
            },
        )

    def test_cat_widths(self):
        self.assertEqual(pyckett.cat_widths_from_quanta(), pyckett.cat_widths)
        self.assertEqual(
            pyckett.cat_widths_from_quanta(), pyckett.cat_widths_from_quanta(6)
        )

        self.assertEqual(
            pyckett.cat_widths_from_quanta(3),
            {
                "x": 13,
                "error": 8,
                "y": 8,
                "degfreed": 2,
                "elower": 10,
                "usd": 3,
                "tag": 7,
                "qnfmt": 4,
                **{f"qn{ul}{i+1}": 2 for ul in ("u", "l") for i in range(6)},
            },
        )

        self.assertEqual(
            pyckett.cat_widths_from_quanta(6),
            {
                "x": 13,
                "error": 8,
                "y": 8,
                "degfreed": 2,
                "elower": 10,
                "usd": 3,
                "tag": 7,
                "qnfmt": 4,
                **{f"qn{ul}{i+1}": 2 for ul in ("u", "l") for i in range(6)},
            },
        )

        self.assertEqual(
            pyckett.cat_widths_from_quanta(8),
            {
                "x": 13,
                "error": 8,
                "y": 8,
                "degfreed": 2,
                "elower": 10,
                "usd": 3,
                "tag": 7,
                "qnfmt": 4,
                **{f"qn{ul}{i+1}": 2 for ul in ("u", "l") for i in range(8)},
            },
        )


class TestLinFormat(unittest.TestCase):
    def test_lin_dtypes(self):
        self.assertEqual(pyckett.lin_dtypes_from_quanta(), pyckett.lin_dtypes)
        self.assertEqual(
            pyckett.lin_dtypes_from_quanta(), pyckett.lin_dtypes_from_quanta(6)
        )

        self.assertEqual(
            pyckett.lin_dtypes_from_quanta(3),
            {
                **{
                    f"qn{ul}{i+1}": pyckett.pickett_int
                    for ul in ("u", "l")
                    for i in range(6)
                },
                "x": np.float64,
                "error": np.float64,
                "weight": np.float64,
                "comment": str,
            },
        )

        self.assertEqual(
            pyckett.lin_dtypes_from_quanta(6),
            {
                **{
                    f"qn{ul}{i+1}": pyckett.pickett_int
                    for ul in ("u", "l")
                    for i in range(6)
                },
                "x": np.float64,
                "error": np.float64,
                "weight": np.float64,
                "comment": str,
            },
        )

        self.assertEqual(
            pyckett.lin_dtypes_from_quanta(8),
            {
                **{
                    f"qn{ul}{i+1}": pyckett.pickett_int
                    for ul in ("u", "l")
                    for i in range(8)
                },
                "x": np.float64,
                "error": np.float64,
                "weight": np.float64,
                "comment": str,
            },
        )


class TestEgyFormat(unittest.TestCase):
    def test_egy_dtypes(self):
        self.assertEqual(pyckett.egy_dtypes_from_quanta(), pyckett.egy_dtypes)
        self.assertEqual(
            pyckett.egy_dtypes_from_quanta(), pyckett.egy_dtypes_from_quanta(6)
        )

        self.assertEqual(
            pyckett.egy_dtypes_from_quanta(3),
            {
                "iblk": np.int64,
                "indx": np.int64,
                "egy": np.float64,
                "err": np.float64,
                "pmix": np.float64,
                "we": np.int64,
                ":": str,
                **{f"qn{i+1}": pyckett.pickett_int for i in range(6)},
            },
        )

        self.assertEqual(
            pyckett.egy_dtypes_from_quanta(6),
            {
                "iblk": np.int64,
                "indx": np.int64,
                "egy": np.float64,
                "err": np.float64,
                "pmix": np.float64,
                "we": np.int64,
                ":": str,
                **{f"qn{i+1}": pyckett.pickett_int for i in range(6)},
            },
        )

        self.assertEqual(
            pyckett.egy_dtypes_from_quanta(8),
            {
                "iblk": np.int64,
                "indx": np.int64,
                "egy": np.float64,
                "err": np.float64,
                "pmix": np.float64,
                "we": np.int64,
                ":": str,
                **{f"qn{i+1}": pyckett.pickett_int for i in range(8)},
            },
        )

    def test_egy_widths(self):
        self.assertEqual(pyckett.egy_widths_from_quanta(), pyckett.egy_widths)
        self.assertEqual(
            pyckett.egy_widths_from_quanta(), pyckett.egy_widths_from_quanta(6)
        )

        self.assertEqual(
            pyckett.egy_widths_from_quanta(3),
            {
                "iblk": 6,
                "indx": 5,
                "egy": 18,
                "err": 18,
                "pmix": 11,
                "we": 5,
                ":": 1,
                **{f"qn{i+1}": 3 for i in range(6)},
            },
        )

        self.assertEqual(
            pyckett.egy_widths_from_quanta(6),
            {
                "iblk": 6,
                "indx": 5,
                "egy": 18,
                "err": 18,
                "pmix": 11,
                "we": 5,
                ":": 1,
                **{f"qn{i+1}": 3 for i in range(6)},
            },
        )

        self.assertEqual(
            pyckett.egy_widths_from_quanta(8),
            {
                "iblk": 6,
                "indx": 5,
                "egy": 18,
                "err": 18,
                "pmix": 11,
                "we": 5,
                ":": 1,
                **{f"qn{i+1}": 3 for i in range(8)},
            },
        )


class TestFileFormats(unittest.TestCase):
    def test_mecn(self):
        lin = pyckett.lin_to_df("tests/resources/MeCN.lin", sort=False)
        par = pyckett.parvar_to_dict("tests/resources/MeCN.par")
        int = pyckett.int_to_dict("tests/resources/MeCN.int")

        active_qns = {
            "qnu1": True,
            "qnu2": True,
            "qnu3": True,
            "qnu4": True,
            "qnu5": False,
            "qnu6": False,
            "qnl1": True,
            "qnl2": True,
            "qnl3": True,
            "qnl4": True,
            "qnl5": False,
            "qnl6": False,
        }
        self.assertEqual(pyckett.get_active_qns(lin), active_qns)

        vib_digits = 1
        self.assertEqual(pyckett.get_vib_digits(par), vib_digits)

        all_states = 9
        self.assertEqual(pyckett.get_all_states(vib_digits), all_states)

        results = pyckett.run_spfit_v(par, lin)
        parsed_result = pyckett.parse_fit_result(results["msg"], results["var"])

        rms = 0.018262
        self.assertTrue(np.isclose(parsed_result["rms"], rms, rtol=1e-3))

        wrms = 0.89593
        self.assertTrue(np.isclose(float(parsed_result["wrms"]), wrms, rtol=1e-3))

        rejected_lines = 0
        self.assertEqual(parsed_result["rejected_lines"], rejected_lines)

        diverging = "NEVER"
        self.assertEqual(parsed_result["diverging"], diverging)

        var = results["var"]
        results = pyckett.run_spcat_v(var, int)

        cat = results["cat"]
        egy = results["egy"]

        query = "qnu1 == 1 and qnu2 == 0 and qnu3 == 0 and qnu4 == 1\
		and qnl1 == 0 and qnl2 == 0 and qnl3 == 0 and qnl4 == 0"
        view = cat.query(query)
        freq = view["x"].values[0]

        assert_freq = 18397.783
        self.assertTrue(np.isclose(freq, assert_freq))

        query = "qn1 == 60 and qn2 == 1 and qn3 == 2 and qn4 == 60"
        view = egy.query(query)
        egy = view["egy"].values[0]

        assert_egy = 4.966675
        self.assertTrue(np.isclose(egy, assert_egy))

    def test_CH2O(self):
        lin = pyckett.lin_to_df("tests/resources/ch2o.lin", sort=False)
        par = pyckett.parvar_to_dict("tests/resources/ch2o.par")
        int = pyckett.int_to_dict("tests/resources/ch2o.int")

        active_qns = {
            "qnu1": True,
            "qnu2": True,
            "qnu3": True,
            "qnu4": False,
            "qnu5": False,
            "qnu6": False,
            "qnl1": True,
            "qnl2": True,
            "qnl3": True,
            "qnl4": False,
            "qnl5": False,
            "qnl6": False,
        }
        self.assertEqual(pyckett.get_active_qns(lin), active_qns)

        vib_digits = 1
        self.assertEqual(pyckett.get_vib_digits(par), vib_digits)

        all_states = 9
        self.assertEqual(pyckett.get_all_states(vib_digits), all_states)

        results = pyckett.run_spfit_v(par, lin)
        parsed_result = pyckett.parse_fit_result(results["msg"], results["var"])

        rms = 0.199064
        self.assertTrue(np.isclose(parsed_result["rms"], rms, rtol=1e-3))

        wrms = 0.56044
        self.assertTrue(np.isclose(float(parsed_result["wrms"]), wrms, rtol=1e-3))

        rejected_lines = 0
        self.assertEqual(parsed_result["rejected_lines"], rejected_lines)

        diverging = "NEVER"
        self.assertEqual(parsed_result["diverging"], diverging)

        var = results["var"]
        results = pyckett.run_spcat_v(var, int)

        cat = results["cat"]
        egy = results["egy"]

        query = "qnu1 == 37 and qnu2 == 7 and qnu3 == 30\
		and qnl1 == 37 and qnl2 == 7 and qnl3 == 31"
        view = cat.query(query)
        freq = view["x"].values[0]

        assert_freq = 138.7619
        self.assertTrue(np.isclose(freq, assert_freq))

        query = "qn1 == 4 and qn2 == 0 and qn3 == 4"
        view = egy.query(query)
        egy = view["egy"].values[0]

        assert_egy = 24.259666
        self.assertTrue(np.isclose(egy, assert_egy))

    def test_CH2O_more_quanta(self):
        pyckett.QUANTA = 7
        try:

            lin = pyckett.lin_to_df("tests/resources/ch2o.lin", sort=False)
            par = pyckett.parvar_to_dict("tests/resources/ch2o.par")
            int = pyckett.int_to_dict("tests/resources/ch2o.int")

            par["SPIND"] = 3333
            par["NLINE"] = -abs(par["NLINE"])
            int["FEND"] = 5

            results = pyckett.run_spcat_v(par, int)

            cat = results["cat"]
            lin = cat.query("y > 1E-6").copy()

            lin["error"] = 0.050
            lin["weight"] = 1
            lin["comment"] = ""

            np.random.seed(112)
            lin["x"] = lin["x"] + np.random.rand(len(lin)) * 0.1

            results_fit = pyckett.run_spfit_v(par, lin)
            parsed_result = pyckett.parse_fit_result(
                results_fit["msg"], results_fit["var"]
            )

            results = pyckett.run_spfit_v(par, lin)
            parsed_result = pyckett.parse_fit_result(results["msg"], results["var"])

            rms = 0.061034
            self.assertTrue(np.isclose(parsed_result["rms"], rms, rtol=1e-3))

            wrms = 0.59024
            self.assertTrue(np.isclose(float(parsed_result["wrms"]), wrms, rtol=1e-3))

            rejected_lines = 0
            self.assertEqual(parsed_result["rejected_lines"], rejected_lines)

            diverging = "LAST"
            self.assertEqual(parsed_result["diverging"], diverging)

        except Exception as E:
            raise
        finally:
            pyckett.QUANTA = 6

    # @Luis: Ask Marie-Aline for some test data for a more than six quanta example


# - test add parameter
# - test omit parameter
# - test finalize
# - test get_dr_candidates
# - test CLI tools

# - think about running flake8 and black


# Run python -m unittest discover from main directory
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Author: Luis Bonah
# Description : CLI tool for showing parameter uncertainties

import os
import argparse
import pyckett

def duplicates():
    parser = argparse.ArgumentParser(prog="Remove duplicates from *.lin file")

    parser.add_argument("linfile", type=str, help="Filename of the .lin file")
    parser.add_argument(
        "--keep",
        type=str,
        nargs=1,
        help="Which values to keep ('first' or 'last')",
        default='last',
    )

    args = parser.parse_args()


    linfname = args.linfile
    keep = args.keep

    try:
        lin = pyckett.lin_to_df(linfname, sort=False)
    except FileNotFoundError:
        linfname = str(linfname) + '.lin'
        lin = pyckett.lin_to_df(linfname, sort=False)
    
    if '.lin' in linfname:
        duplicatesfname = linfname.replace('.lin', '_duplicates.lin')
    else:
        duplicatesfname = f'{linfname}_duplicates.lin'


    qn_labels = [f'qn{ul}{i+1}' for ul in 'ul' for i in range(pyckett.QUANTA)]
    duplicates = lin[lin.duplicated(subset=qn_labels, keep=keep)]
    n_duplicates = len(duplicates)

    with open(duplicatesfname, 'w+') as file:
        file.write(pyckett.df_to_lin(duplicates))
    
    with open(linfname, 'w+') as file:
        file.write(pyckett.df_to_lin(lin.drop_duplicates(subset=qn_labels, keep=keep)))


    print(f'Removed {n_duplicates} duplicates from your *.lin file.')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Author: Luis Bonah
# Description : CLI tool for checking which parameter to omit from fit

import os.path
import argparse
import pyckett


def omitparameters():
    parser = argparse.ArgumentParser(prog="Omit Parameters")

    parser.add_argument("linfile", type=str, help="Filename of the .lin file")
    parser.add_argument(
        "parfile", type=str, nargs="?", help="Filename of the .par file"
    )

    parser.add_argument(
        "--skipinterstate",
        action="store_true",
        help="Do not test interstate parameters",
    )
    parser.add_argument(
        "--skiprotational",
        action="store_true",
        help="Do not test pure rotational parameters",
    )
    parser.add_argument(
        "--skipfixed", action="store_true", help="Do not test fixed parameters"
    )
    parser.add_argument(
        "--skipglobal", action="store_true", help="Do not test global parameters"
    )

    parser.add_argument(
        "--skipparupdate",
        action="store_true",
        help="Do not reset .par parametes NPAR, NLINE, THRESH",
    )
    parser.add_argument(
        "--sortwrms",
        action="store_true",
        help="Sort results by their WRMS values instead of their RMS value",
    )

    args = parser.parse_args()

    linfname = args.linfile
    base, ext = os.path.splitext(linfname)
    if not ext:
        linfname = linfname + ".lin"

    lin = pyckett.lin_to_df(linfname, sort=False)
    par = pyckett.parvar_to_dict(
        args.parfile if args.parfile else linfname.replace(".lin", ".par")
    )

    VIB_DIGITS = pyckett.get_vib_digits(par)
    ALL_STATES = pyckett.get_all_states(VIB_DIGITS)

    if not args.skipparupdate:
        par.update(pyckett.PARUPDATE)

    kwargs = {
        "VIB_DIGITS": VIB_DIGITS,
        "ALL_STATES": ALL_STATES,
        "skipglobal": args.skipglobal,
        "skipfixed": args.skipfixed,
        "skipinterstate": args.skipinterstate,
        "skiprotational": args.skiprotational,
        "sort_by_wrms": args.sortwrms,
    }

    omitparameters_core(par, lin, **kwargs)


def omitparameters_core(
    par,
    lin,
    VIB_DIGITS=1,
    ALL_STATES=9,
    skipglobal=False,
    skipfixed=True,
    skipinterstate=True,
    skiprotational=True,
    report=True,
    sort_by_wrms=False,
):
    candidates = ["INITIAL"]

    for param in par["PARAMS"]:
        id = param[0]
        parsed_id = pyckett.parse_param_id(id, VIB_DIGITS)

        isinterstate = parsed_id["v1"] != parsed_id["v2"]
        isrotational = parsed_id["v1"] == parsed_id["v2"]
        isglobal = parsed_id["v1"] == parsed_id["v2"] == ALL_STATES
        isfixed = abs(param[2]) < pyckett.ZEROTHRESHOLD

        if skipfixed and isfixed:
            continue

        if skipglobal and isglobal:
            continue

        if skipinterstate and isinterstate:
            continue

        if skiprotational and isrotational:
            continue

        candidates.append(id)

    runs = pyckett.omit_parameter(par, lin, candidates, sort_by_wrms=sort_by_wrms)

    if report:
        header = "         ID    |    RMS [kHz] |   WRMS | RejLines |  Diverging "
        print(header)
        print("-" * len(header))
        for results in runs:
            if results["rms"] is None:
                print(
                    f"ID {results['id']:14} | FAILED! This parameter could be essential."
                )
            else:
                id, rms, wrms, rejected_lines, diverging = (
                    results["id"],
                    results["rms"] * 1000,
                    results["wrms"],
                    results["stats"]["rejected_lines"],
                    results["stats"]["diverging"],
                )
                print(
                    f"{id:14} | {rms:12.2f} | {wrms:6.2f} | {rejected_lines:8.0f} | {diverging:10} "
                )
                # print(f"ID {results['id']:8}; RMS of {results['rms']*1000 :10.2f} kHz; Rejected lines {results['stats']['rejected_lines']:8}; Diverging {results['stats']['diverging']:8};")

    return runs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Author: Luis Bonah
# Description : CLI tool for checking which parameter to add to fit

import os.path
import pyckett
import argparse
from concurrent.futures import ThreadPoolExecutor

# @Luis: Provide option for symmetric, linear molecules
# @Luis: Maybe infer from Kmin and Kmax


def addparameters():
    parser = argparse.ArgumentParser(prog="Add Parameters")

    parser.add_argument("linfile", type=str, help="Filename of the .lin file")
    parser.add_argument(
        "parfile", type=str, nargs="?", help="Filename of the .par file"
    )

    parser.add_argument(
        "--skipinterstate",
        action="store_true",
        help="Do not test interstate parameters",
    )
    parser.add_argument(
        "--skiprotational",
        action="store_true",
        help="Do not test pure rotational parameters",
    )
    parser.add_argument(
        "--skipfixed", action="store_true", help="Do not test fixed parameters"
    )
    parser.add_argument(
        "--skipglobal", action="store_true", help="Do not test global parameters"
    )

    parser.add_argument(
        "--skipparupdate",
        action="store_true",
        help="Do not reset .par parametes NPAR, NLINE, THRESH",
    )

    parser.add_argument(
        "--newinteraction",
        type=int,
        nargs="+",
        help="States to find new interactions for",
    )
    parser.add_argument(
        "--initialvalues",
        nargs="*",
        type=float,
        help="Initial values to test for parameters (default 1E-37)",
    )
    parser.add_argument(
        "--skipsavebest",
        action="store_true",
        help='Do not save the best run to "best.par"',
    )
    parser.add_argument(
        "--stateqn", default=4, type=int, help="Quantum number index for state"
    )

    parser.add_argument(
        "--areduction",
        dest="parameters",
        action="store_const",
        const="a_reduction",
        help="Use A-Reduction parameters",
    )
    parser.add_argument(
        "--sreduction",
        dest="parameters",
        action="store_const",
        const="s_reduction",
        help="Use S-Reduction parameters",
    )
    parser.add_argument(
        "--linear",
        dest="parameters",
        action="store_const",
        const="linear",
        help="Use linear parameters",
    )

    parser.add_argument(
        "--paramids",
        type=int,
        nargs="+",
        help="Manually specify the to-be-tested parameter IDs",
    )
    parser.add_argument(
        "--overwritepar",
        action="store_true",
        help="Overwrite the .par file with the best fit",
    )
    parser.add_argument(
        "--sortwrms",
        action="store_true",
        help="Sort results by their WRMS values instead of their RMS value",
    )

    args = parser.parse_args()

    linfname = args.linfile
    base, ext = os.path.splitext(linfname)
    if not ext:
        linfname = linfname + ".lin"

    parfname = args.parfile if args.parfile else linfname.replace(".lin", ".par")

    lin = pyckett.lin_to_df(linfname, sort=False)
    par = pyckett.parvar_to_dict(parfname)

    if not args.skipparupdate:
        par.update(pyckett.PARUPDATE)

    VIB_DIGITS = pyckett.get_vib_digits(par)
    ALL_STATES = pyckett.get_all_states(VIB_DIGITS)

    qnu, qnl = f"qnu{args.stateqn:1.0f}", f"qnl{args.stateqn:1.0f}"

    bestparfile = "best.par"
    if args.skipsavebest:
        bestparfile = None
    if args.overwritepar:
        bestparfile = parfname

    kwargs = {
        "VIB_DIGITS": VIB_DIGITS,
        "ALL_STATES": ALL_STATES,
        "skipglobal": args.skipglobal,
        "skipfixed": args.skipfixed,
        "skipinterstate": args.skipinterstate,
        "skiprotational": args.skiprotational,
        "qnu": qnu,
        "qnl": qnl,
        "parameters": args.parameters,
        "newinteraction": args.newinteraction,
        "initialvalues": args.initialvalues,
        "parameters_to_test": args.paramids,
        "bestparfile": bestparfile,
        "sort_by_wrms": args.sortwrms,
    }

    addparameters_core(par, lin, **kwargs)


def prit(condition, text=""):
    if condition:
        print(text)


def addparameters_core(
    par,
    lin,
    VIB_DIGITS=1,
    ALL_STATES=9,
    qnu="qnu4",
    qnl="qnl4",
    parameters=None,
    skipfixed=False,
    skipinterstate=False,
    skiprotational=False,
    skipglobal=False,
    newinteraction=None,
    initialvalues=None,
    bestparfile=None,
    parameters_to_test=None,
    report=True,
    sort_by_wrms=False,
):
    if not initialvalues:
        initialvalues = [1e-37]

    # Case of manually specified parameter list
    if parameters_to_test:
        candidates = [(id, "") for id in parameters_to_test]

    # Case of infering parameter to test
    else:
        INTERSTATE_PARAMS = pyckett.POSSIBLE_PARAMS_INTERACTION

        if parameters == "linear":
            ROTATIONAL_PARAMS = pyckett.POSSIBLE_PARAMS_LINEAR
        elif parameters == "s_reduction":
            ROTATIONAL_PARAMS = pyckett.POSSIBLE_PARAMS_S
        else:
            ROTATIONAL_PARAMS = pyckett.POSSIBLE_PARAMS_A

        rotational_cands = set()
        interstate_cands = set()
        present_params = set()

        interstates = set()
        rotationstates = set()

        for param in par["PARAMS"]:
            isfixed = abs(param[2]) < pyckett.ZEROTHRESHOLD
            if isfixed and skipfixed:
                continue
            id = param[0]
            present_params.add(id)

            param_dict = pyckett.parse_param_id(id, VIB_DIGITS)
            base_id = pyckett.format_param_id(param_dict, 0)

            isrotational = param_dict["v1"] == param_dict["v2"]
            ud = {"v1": param_dict["v1"], "v2": param_dict["v2"]}

            if skipglobal and param_dict["v1"] == param_dict["v2"] == ALL_STATES:
                continue

            if isrotational:
                params = ROTATIONAL_PARAMS
                cands = rotational_cands
                rotationstates.add(ud["v1"])

            else:
                params = INTERSTATE_PARAMS
                cands = interstate_cands
                interstates.add((ud["v2"], ud["v1"]))

            comment, next_params_ids = params.get(base_id, (None, None))
            if next_params_ids is None:
                prit(report, f"Did not find any candidates for parameter {id}.")
                continue

            next_params_dicts = [
                {**ud, **pyckett.parse_param_id(id, 0)} for id in next_params_ids
            ]
            cands.update(
                [
                    pyckett.format_param_id(dict_, VIB_DIGITS)
                    for dict_ in next_params_dicts
                ]
            )

        if par["NVIB"] < 2:
            lin_states = set()
        else:
            lin_states = set(lin[qnu].unique()) & set(lin[qnl].unique())
        update_base_states = lambda ud, bs: [
            pyckett.format_param_id({**ud, **pyckett.parse_param_id(id, 0)}, VIB_DIGITS)
            for id in bs
        ]

        # Add base parameters for all rotational states
        for v in rotationstates | lin_states:
            ud = {"v1": v, "v2": v}
            rotational_cands.update(
                update_base_states(ud, pyckett.INITIAL_PARAMS_ROTATION)
            )

        # Add base parameters for all interaction states
        if newinteraction:
            interstates.update(
                [
                    tuple(sorted((v1, v2)))
                    for v1 in newinteraction
                    for v2 in newinteraction
                    if v1 != v2
                ]
            )

        for v2, v1 in interstates:
            ud = {"v1": v1, "v2": v2}
            interstate_cands.update(
                update_base_states(ud, pyckett.INITIAL_PARAMS_INTERACTION)
            )

        rotational_cands = rotational_cands - present_params
        get_comment = lambda id: ROTATIONAL_PARAMS.get(
            pyckett.format_param_id(pyckett.parse_param_id(id, VIB_DIGITS), 0), ("",)
        )[0]
        rotational_cands = [(id, get_comment(id)) for id in rotational_cands]

        interstate_cands = interstate_cands - present_params
        get_comment = lambda id: INTERSTATE_PARAMS.get(
            pyckett.format_param_id(pyckett.parse_param_id(id, VIB_DIGITS), 0), ("",)
        )[0]
        interstate_cands = [(id, get_comment(id)) for id in interstate_cands]

        candidates = []
        if not skiprotational:
            candidates.extend(rotational_cands)
        if not skipinterstate:
            candidates.extend(interstate_cands)

    # Set initial values
    tmp = []
    for id, comment in candidates:
        tmp.extend([[[id, initvalue, 1e37, comment]] for initvalue in initialvalues])
    candidates = tmp

    # Calculate initial stats
    init_stats = pyckett.add_parameter(par, lin, [[]], sort=False)[0]
    init_stats["id"] = ["INITIAL"]
    results = pyckett.add_parameter(
        par, lin, candidates, sort=True, sort_by_wrms=sort_by_wrms
    )
    sort_key = "wrms" if sort_by_wrms else "rms"

    # Test new candidate parameters
    header = "         ID    |    RMS [kHz] |   WRMS | RejLines |  Diverging | Init Value | Final Value "
    prit(report, "")
    prit(report, header)
    prit(report, "-" * len(header))

    best_stats = init_stats
    for i, stats in enumerate(results):
        if (
            stats[sort_key] < best_stats[sort_key]
            and stats["stats"]["rejected_lines"]
            <= best_stats["stats"]["rejected_lines"]
            and stats["stats"]["diverging"] != "LAST"
        ):
            best_stats = stats

        id = stats["id"][0]
        rms = stats["rms"] * 1000
        wrms = stats["stats"]["wrms"]
        initial = stats["params"][-1][1]
        final = stats["par"][-1][1]
        rejected_lines = stats["stats"]["rejected_lines"]
        diverging = stats["stats"]["diverging"]

        prit(
            report,
            f"{id:14} | {rms:12.2f} | {wrms:6.2f} | {rejected_lines:8.0f} | {diverging:10} | {initial:10.2e} | {final:10.2e} ",
        )

    # Final report
    init_value = f'{init_stats[sort_key]*1000 :.2f} kHz' if sort_key == 'rms' else f'{init_stats[sort_key]:.2f}'
    prit(
        report,
        f'\nInitial values were an {sort_key.upper()} of {init_value}, {init_stats["stats"]["rejected_lines"]} rejected lines, and diverging {init_stats["stats"]["diverging"]}.',
    )
    best_value = f'{best_stats[sort_key]*1000 :.2f} kHz' if sort_key == 'rms' else f'{best_stats[sort_key]:.2f}'
    prit(
        report,
        f'\nBest run is parameter {best_stats["id"][0]} with a final parameter value of {best_stats["par"][-1][1]:.2e}, {sort_key.upper()} of {best_value}, {best_stats["stats"]["rejected_lines"]} rejected lines, and diverging {best_stats["stats"]["diverging"]}.',
    )

    if bestparfile is not None:
        with open(bestparfile, "w+") as file:
            save_par = par.copy()
            save_par["PARAMS"] = best_stats["par"]
            file.write(pyckett.dict_to_parvar(save_par))
        prit(report, '\nWrote the best run to "best.par".')
    return (init_stats, best_stats, results)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Author: Luis Bonah
# Description : CLI tool for showing parameter uncertainties

import argparse
import pyckett


def pmix():
    parser = argparse.ArgumentParser(prog="Plot mixing coefficients from .egy file")

    parser.add_argument("egyfile", type=str, help="Filename of the .egy file")
    parser.add_argument("queries", nargs="+", help="Queries for plots")

    args = parser.parse_args()

    egy = pyckett.egy_to_df(args.egyfile)

    pyckett.mixing_coefficient(egy, args.queries)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Author: Luis Bonah
# Description : CLI tool to run SPFIT/SPCAT

import os
import subprocess
import argparse

subprocess_kwargs = {
    "stdout": subprocess.PIPE,
    "stderr": subprocess.PIPE,
}


def runfit():
    parser = argparse.ArgumentParser(prog="Run SPFIT")
    parser.add_argument(
        "filenames", type=str, nargs="+", help="Filenames to pass to SPFIT"
    )
    args = parser.parse_args()

    path = os.environ.get("PYCKETT_SPFIT_PATH", "spfit")
    command = [path, *args.filenames]

    with subprocess.Popen(command, **subprocess_kwargs) as process:
        while process.poll() is None:
            text = process.stdout.read1().decode("utf-8")
            print(text, end="", flush=True)

        text = process.stdout.read().decode("utf-8")
        print(text)


def runpredictions():
    parser = argparse.ArgumentParser(prog="Run SPCAT")
    parser.add_argument(
        "filenames", type=str, nargs="+", help="Filenames to pass to SPCAT"
    )
    args = parser.parse_args()

    path = os.environ.get("PYCKETT_SPCAT_PATH", "spcat")
    command = [path, *args.filenames]

    with subprocess.Popen(command, **subprocess_kwargs) as process:
        while process.poll() is None:
            text = process.stdout.read1().decode("utf-8")
            print(text, end="", flush=True)

        text = process.stdout.read().decode("utf-8")
        print(text)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Author: Luis Bonah
# Description : CLI tool to compute partition function of egy file

import pyckett
import numpy as np
import argparse


recommended_use = """
Recommended procedure for determining the partition function

Determine the partition function for the ground state unless all vibrational states are treated correctly.
Set Jmax and Kmax sufficiently high for convergence but not too high (too avoid rollover). When in doubt check with the --convergence option.
"""

Temperatures = [
    300.000,
    225.000,
    150.000,
    75.000,
    37.500,
    18.750,
    9.37,
]

k = 1.380649e-23
h = 6.62607015e-34
c = 299792458

cm1_to_Joules = h * c * 100
factor = cm1_to_Joules / k

# This is the factor that is used in Pickett's spcat procedure
# The factor can be found in the file calcat.c in line 97
factor_pickett = 1.43878

# Uncomment the following line to check if the result fits to the spcat value
# factor = factor_pickett


def partitionfunction_at_temperature(egy_df, T, factor=factor):
    egy_df["tmp"] = (2 * egy_df["qn1"] + 1) * np.exp(-egy_df["egy"] * factor / T)
    return egy_df["tmp"].sum()


def partitionfunction():
    parser = argparse.ArgumentParser(
        prog="Partitionfunction",
        epilog=recommended_use,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument("egyfile", type=str, help="Filename of the .egy file")
    parser.add_argument(
        "--temperatures",
        "-T",
        metavar="T",
        nargs="*",
        type=float,
        help="Additional temperatures in Kelvin to evaluate the partition function for",
    )
    parser.add_argument(
        "--pickett",
        "-p",
        dest="use_pickett_factor",
        action="store_true",
        help="Use the (less precise) numeric values from SPCAT",
    )
    parser.add_argument(
        "--convergence",
        "-c",
        dest="convergence",
        action="store_true",
        help="Produce convergence plot",
    )

    args = parser.parse_args()

    egyfilename = args.egyfile

    if args.temperatures:
        Temperatures.extend(args.temperatures)
        Temperatures.sort(reverse=True)

    egy_df = pyckett.egy_to_df(egyfilename)
    calc_partition_function(
        egy_df,
        Temperatures,
        convergence_plot=args.convergence,
        print_output=True,
        use_pickett_factor=args.use_pickett_factor,
    )


def calc_partition_function(
    egy_df,
    temperatures,
    print_output=False,
    convergence_plot=False,
    use_pickett_factor=False,
):
    Jmax = egy_df["qn1"].max()
    Kamax = egy_df["qn2"].max()
    Kcmax = egy_df["qn3"].max()

    tmp_factor = factor_pickett if use_pickett_factor else factor

    partition_functions = {
        T: partitionfunction_at_temperature(egy_df, T, factor=tmp_factor)
        for T in temperatures
    }

    if print_output:
        print(
            f"Partition function with J_max = {Jmax}, Ka_max = {Kamax}, Kc_max = {Kcmax}"
        )
        for key, value in partition_functions.items():
            print(f"T = {key:6.2f}; Q = {value:12.4f}; log Q = {np.log10(value):7.4f}")

    if convergence_plot:
        import matplotlib.pyplot as plt

        Js = np.arange(0, Jmax)
        results = np.zeros((len(temperatures), len(Js)))

        for i, Jupper in enumerate(Js):
            tmp_df = egy_df.loc[egy_df["qn1"] <= Jupper].copy()

            for j, T in enumerate(temperatures):
                pf = partitionfunction_at_temperature(tmp_df, T, factor=tmp_factor)
                results[j, i] = pf

        for j, T in enumerate(temperatures):
            ys = results[j, :]
            plt.plot(Js, ys, label=f"{T=} K")

        plt.xlabel(r"$J_{max}$")
        plt.ylabel("$Q$")
        plt.legend()
        plt.show()

    return partition_functions



from .shortcuts import *
from .addparameters import *
from .uncertainties import *
from .omitparameters import *
from .partitionfunction import *
from .separatefits import *
from .automaticfit import *
from .pmix import *
from .report import *
from .duplicates import *
